import os
import shutil
import socket
import tempfile
import time
from os.path import join
from uuid import uuid4

from robot.libraries.BuiltIn import BuiltIn
from SeleniumLibrary.base import LibraryComponent, keyword
from six.moves.urllib.request import urlopen


class ServerKeywords(LibraryComponent):
    _handles = []
    _tmpdirs = {}
    _notebook_dirs = {}
    _ports = {}
    _base_urls = {}
    _tokens = {}

    @keyword
    def start_new_jupyter_server(
        self,
        command=None,
        port=None,
        base_url=None,
        notebook_dir=None,
        token=None,
        *args,
        **config,
    ):
        """ Start a Jupyter server

            If not configured, the HOME environment variable and current
            working directory will be set to avoid leaking configuration
            between between runs (or the test instance) itself. These
            directories will be cleaned up after the server process is
            terminated.
        """
        command = command or "jupyter"
        port = port or self.get_unused_port()
        base_url = base_url or "/@rf/"
        token = str(uuid4()) if token is None else token

        BuiltIn().import_library("Process")
        plib = BuiltIn().get_library_instance("Process")

        tmpdir = tempfile.mkdtemp()

        if "env:HOME" not in config:
            home_dir = join(tmpdir, "home")
            os.mkdir(home_dir)
            config["env:HOME"] = home_dir

        if "stdout" not in config:
            config["stdout"] = join(tmpdir, "server.log")

        if "stderr" not in config:
            config["stderr"] = "STDOUT"

        if notebook_dir is None:
            notebook_dir = join(tmpdir, "notebooks")
            os.mkdir(notebook_dir)
            config["cwd"] = notebook_dir

        args = args or self.build_jupyter_server_arguments(port, base_url, token)

        handle = plib.start_process(command, *args, **config)

        self._handles += [handle]
        self._tmpdirs[handle] = tmpdir
        self._notebook_dirs[handle] = notebook_dir
        self._ports[handle] = port
        self._base_urls[handle] = base_url
        self._tokens[handle] = token

        return handle

    @keyword
    def build_jupyter_server_arguments(self, port, base_url, token):
        """ Some default jupyter arguments
        """
        return [
            "notebook",
            "--no-browser",
            "--debug",
            "--port={}".format(port),
            "--NotebookApp.token='{}'".format(token),
            "--NotebookApp.base_url='{}'".format(base_url),
        ]

    @keyword
    def copy_files_to_jupyter_directory(self, *sources, **kwargs):
        """ Copy some files into the (temporary) jupyter server root
        """
        nbserver = kwargs.get("nbserver", self._handles[-1])
        notebook_dir = self._notebook_dirs[nbserver]
        BuiltIn().import_library("OperatingSystem")
        osli = BuiltIn().get_library_instance("OperatingSystem")
        osli.copy_files(*(list(sources) + [notebook_dir]))

    @keyword
    def copy_files_from_jupyter_directory(self, *src_and_dest, **kwargs):
        """ Copy some files from the (temporary) jupyter server root

            Patterns will have the notebook directory prepended
        """
        nbserver = kwargs.get("nbserver", self._handles[-1])
        notebook_dir = self._notebook_dirs[nbserver]
        BuiltIn().import_library("OperatingSystem")
        osli = BuiltIn().get_library_instance("OperatingSystem")
        sources = [join(notebook_dir, src) for src in src_and_dest[:-1]]
        dest = src_and_dest[-1]
        osli.copy_files(*sources + [dest])

    @keyword
    def get_jupyter_directory(self, nbserver=None):
        nbserver = nbserver if nbserver is not None else self._handles[-1]
        return self._notebook_dirs[nbserver]

    @keyword
    def wait_for_jupyter_server_to_be_ready(self, *nbservers, **kwargs):
        """  Wait for the most-recently started Jupyter server to be ready
        """
        interval = float(kwargs.get("interval", 0.5))
        retries = int(kwargs.get("retries", 60))

        if not nbservers:
            if not self._handles:
                return 0
            nbservers = [self._handles[-1]]

        ready = 0
        error = None

        while retries and ready != len(nbservers):
            retries -= 1
            ready = 0
            try:
                for nbh in nbservers:
                    urlopen(self.get_jupyter_server_url(nbh))
                    ready += 1
            except Exception as _error:
                time.sleep(interval)
                error = _error

        assert ready == len(
            nbservers
        ), "Only {} of {} servers were ready after {}s. Last error: {} {}".format(
            ready, len(nbservers), interval * retries, type(error), error
        )
        return ready

    @keyword
    def get_jupyter_server_url(self, nbserver=None):
        nbh = nbserver or self._handles[-1]
        return "http://localhost:{}{}".format(self._ports[nbh], self._base_urls[nbh])

    @keyword
    def get_jupyter_server_token(self, nbserver=None):
        nbh = nbserver or self._handles[-1]
        return self._tokens[nbh]

    @keyword
    def wait_for_new_jupyter_server_to_be_ready(self, command=None, *args, **config):
        handle = self.start_new_jupyter_server(command, *args, **config)
        self.wait_for_jupyter_server_to_be_ready(handle)
        return handle

    @keyword
    def terminate_all_jupyter_servers(self):
        """ Close all Jupyter servers started by JupyterLibrary
        """
        plib = BuiltIn().get_library_instance("Process")

        self.wait_for_jupyter_server_to_be_ready()

        terminated = 0
        shutdown = 0
        for nbh in self._handles:
            url = self.get_jupyter_server_url(nbh)
            token = self.get_jupyter_server_token(nbh)
            try:
                urlopen("{}api/shutdown?token={}".format(url, token), data=[])
                shutdown += 1
            except Exception as err:
                BuiltIn().log(err)

        if shutdown:
            for nbh in self._handles:
                try:
                    plib.terminate_process(nbh, kill=True)
                    terminated += 1
                except Exception as err:
                    BuiltIn().log(err)

        # give processes a mo to shutdown
        if terminated or shutdown:
            BuiltIn().sleep("5s")
            for nbh in self._handles:
                shutil.rmtree(self._tmpdirs[nbh])

        self._handles = []
        self._tmpdirs = {}
        self._notebook_dirs = {}
        self._ports = {}
        self._base_urls = {}
        self._tokens = {}

        return terminated

    @keyword
    def get_unused_port(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("localhost", 0))
        s.listen(1)
        port = s.getsockname()[1]
        s.close()
        return port



""" Backport of SeleniumLibrary >3.2.0's `Press Keys`
"""
from robot.utils import plural_or_not

# flake8: noqa
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from SeleniumLibrary.base import LibraryComponent, keyword
from SeleniumLibrary.utils import is_truthy


class KeysKeywords(LibraryComponent):
    @keyword
    def press_keys(self, locator=None, *keys):
        """Simulates user pressing key(s) to an element or on the active browser.
        If ``locator`` evaluates as false, see `Boolean arguments` for more
        details, then the ``keys`` are sent to the currently active browser.
        Otherwise element is searched and ``keys`` are send to the element
        identified by the ``locator``. In later case, keyword fails if element
        is not found. See the `Locating elements` section for details about
        the locator syntax.
        ``keys`` arguments can contain one or many strings, but it can not
        be empty. ``keys`` can also be a combination of
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html|Selenium Keys]
        and strings or a single Selenium Key. If Selenium Key is combined
        with strings, Selenium key and strings must be separated by the
        `+` character, like in `CONTROL+c`. Selenium Keys
        are space and case sensitive and Selenium Keys are not parsed
        inside of the string. Example AALTO, would send string `AALTO`
        and `ALT` not parsed inside of the string. But `A+ALT+O` would
        found Selenium ALT key from the ``keys`` argument. It also possible
        to press many Selenium Keys down at the same time, example
        'ALT+ARROW_DOWN`.
        If Selenium Keys are detected in the ``keys`` argument, keyword
        will press the Selenium Key down, send the strings and
         then release the Selenium Key. If keyword needs to send a Selenium
        Key as a string, then each character must be separated with
        `+` character, example `E+N+D`.
        `CTRL` is alias for
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html#selenium.webdriver.common.keys.Keys.CONTROL|Selenium CONTROL]
        and ESC is alias for
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html#selenium.webdriver.common.keys.Keys.ESCAPE|Selenium ESCAPE]
        New in SeleniumLibrary 3.3
        Examples:
        | `Press Keys` | text_field | AAAAA          |            | # Sends string "AAAAA" to element.                                                |
        | `Press Keys` | None       | BBBBB          |            | # Sends string "BBBBB" to currently active browser.                               |
        | `Press Keys` | text_field | E+N+D          |            | # Sends string "END" to element.                                                  |
        | `Press Keys` | text_field | XXX            | YY         | # Sends strings "XXX" and "YY" to element.                                        |
        | `Press Keys` | text_field | XXX+YY         |            | # Same as above.                                                                  |
        | `Press Keys` | text_field | ALT+ARROW_DOWN |            | # Pressing "ALT" key down, then pressing ARROW_DOWN and then releasing both keys. |
        | `Press Keys` | text_field | ALT            | ARROW_DOWN | # Pressing "ALT" key and then pressing ARROW_DOWN.                                |
        | `Press Keys` | text_field | CTRL+c         |            | # Pressing CTRL key down, sends string "c" and then releases CTRL key.            |
        | `Press Keys` | button     | RETURN         |            | # Pressing "ENTER" key to element.                                                |
        """
        parsed_keys = self._parse_keys(*keys)
        if is_truthy(locator):
            self.info("Sending key(s) %s to %s element." % (keys, locator))
        else:
            self.info("Sending key(s) %s to page." % str(keys))
        self._press_keys(locator, parsed_keys)

    def _map_ascii_key_code_to_key(self, key_code):
        map = {
            0: Keys.NULL,
            8: Keys.BACK_SPACE,
            9: Keys.TAB,
            10: Keys.RETURN,
            13: Keys.ENTER,
            24: Keys.CANCEL,
            27: Keys.ESCAPE,
            32: Keys.SPACE,
            42: Keys.MULTIPLY,
            43: Keys.ADD,
            44: Keys.SEPARATOR,
            45: Keys.SUBTRACT,
            56: Keys.DECIMAL,
            57: Keys.DIVIDE,
            59: Keys.SEMICOLON,
            61: Keys.EQUALS,
            127: Keys.DELETE,
        }
        key = map.get(key_code)
        if key is None:
            key = chr(key_code)
        return key

    def _map_named_key_code_to_special_key(self, key_name):
        try:
            return getattr(Keys, key_name)
        except AttributeError:
            message = "Unknown key named '%s'." % (key_name)
            self.debug(message)
            raise ValueError(message)

    def _press_keys(self, locator, parsed_keys):
        if is_truthy(locator):
            element = self.find_element(locator)
        else:
            element = None
        for parsed_key in parsed_keys:
            actions = ActionChains(self.driver)
            special_keys = []
            for key in parsed_key:
                if self._selenium_keys_has_attr(key.original):
                    special_keys = self._press_keys_special_keys(
                        actions, element, parsed_key, key, special_keys
                    )
                else:
                    self._press_keys_normal_keys(actions, element, key)
            for special_key in special_keys:
                self.info("Releasing special key %s." % special_key.original)
                actions.key_up(special_key.converted)
            actions.perform()

    def _press_keys_normal_keys(self, actions, element, key):
        self.info("Sending key%s %s" % (plural_or_not(key.converted), key.converted))
        if element:
            actions.send_keys_to_element(element, key.converted)
        else:
            actions.send_keys(key.converted)

    def _press_keys_special_keys(self, actions, element, parsed_key, key, special_keys):
        if len(parsed_key) == 1 and element:
            self.info("Pressing special key %s to element." % key.original)
            actions.send_keys_to_element(element, key.converted)
        elif len(parsed_key) == 1 and not element:
            self.info("Pressing special key %s to browser." % key.original)
            actions.send_keys(key.converted)
        else:
            self.info("Pressing special key %s down." % key.original)
            actions.key_down(key.converted)
            special_keys.append(key)
        return special_keys

    def _parse_keys(self, *keys):
        if not keys:
            raise AssertionError('"keys" argument can not be empty.')
        list_keys = []
        for key in keys:
            separate_keys = self._separate_key(key)
            separate_keys = self._convert_special_keys(separate_keys)
            list_keys.append(separate_keys)
        return list_keys

    def _parse_aliases(self, key):
        if key == "CTRL":
            return "CONTROL"
        if key == "ESC":
            return "ESCAPE"
        return key

    def _separate_key(self, key):
        one_key = ""
        list_keys = []
        for char in key:
            if char == "+" and one_key != "":
                list_keys.append(one_key)
                one_key = ""
            else:
                one_key += char
        if one_key:
            list_keys.append(one_key)
        return list_keys

    def _convert_special_keys(self, keys):
        KeysRecord = namedtuple("KeysRecord", "converted, original")
        converted_keys = []
        for key in keys:
            key = self._parse_aliases(key)
            if self._selenium_keys_has_attr(key):
                converted_keys.append(KeysRecord(getattr(Keys, key), key))
            else:
                converted_keys.append(KeysRecord(key, key))
        return converted_keys

    def _selenium_keys_has_attr(self, key):
        try:
            return hasattr(Keys, key)
        except UnicodeError:  # To support Python 2 and non ascii characters.
            return False



from PIL import Image
from robot.libraries.BuiltIn import BuiltIn
from SeleniumLibrary.base import LibraryComponent, keyword


class ScreenshotKeywords(LibraryComponent):
    @keyword
    def capture_element_screenshot(self, locator, filename):
        el = self.find_element(locator)
        bbox = self.normalize_bounding_box({**el.location, **el.size})
        filename = BuiltIn().run_keyword("Capture Page Screenshot", filename)
        self.crop_image(filename, **bbox)

    def normalize_bounding_box(self, bbox):
        """ Normalize a bounding box to reasonable defaults
            TODO: also handle larger-than-screen elements (if that's a thing)
        """
        return {
            k: (
                max(int(round(v)), 1)
                if k in ["width", "height"]
                else max(int(round(v)), 0)
            )
            for k, v in bbox.items()
        }

    @keyword
    def crop_image(self, in_file, x, y, width, height, out_file=None):
        out_file = out_file or in_file
        img = Image.open(in_file)
        area = img.crop((int(x), int(y), int(x + width), int(y + height)))

        with open(out_file, "wb") as output:
            area.save(output, "png")

        return out_file



import os
import re

from setuptools import setup


HERE = os.path.dirname(__file__)

with open(os.path.join(HERE, "src", "JupyterLibrary", "_version.py")) as fp:
    version = re.findall(r"""__version__ = "([^"]+)""", fp.read())[0]

setup(version=version)
